import sys

import pytest
from toy_crypto.ec import Curve
from toy_crypto.nt import Modulus


class TestSeriousCurve:
    """
    Tests involving example curve from Serious Cryptography
    """

    sc_parameters = (-4, 0, Modulus(191))
    sc_generator = (146, 131)
    curve = Curve(*sc_parameters)
    G = curve.point(*sc_generator)
    Px = 3
    Py = 46
    Qx = 146

    # Computed using SageMath
    sage_vectors: dict[int, tuple[int, int]] = {
        1: (146, 131), 2: (8, 174), 3: (137, 161), 4: (4, 96),
        5: (61, 163), 6: (34, 83), 7: (171, 22), 8: (18, 52),
        9: (174, 180), 10: (77, 133), 11: (186, 75), 12: (136, 46),
        13: (29, 4), 14: (48, 57), 15: (155, 61), 16: (158, 26),
        17: (71, 120), 18: (102, 69), 19: (179, 109), 20: (15, 100),
        21: (122, 124), 22: (3, 46), 23: (87, 55), 24: (79, 33),
        25: (74, 114), 26: (10, 177), 27: (53, 110), 28: (32, 172),
        29: (56, 98), 30: (25, 98), 31: (99, 177), 32: (100, 80),
        33: (42, 167), 34: (52, 145), 35: (165, 10), 36: (156, 184),
        37: (110, 98), 38: (160, 79), 39: (82, 177), 40: (98, 121),
        41: (19, 174), 42: (50, 29), 43: (164, 17), 44: (6, 1),
        45: (111, 188), 46: (67, 166), 47: (140, 38), 48: (2, 0),
        49: (140, 153), 50: (67, 25), 51: (111, 3), 52: (6, 190),
        53: (164, 174), 54: (50, 162), 55: (19, 17), 56: (98, 70),
        57: (82, 14), 58: (160, 112), 59: (110, 93), 60: (156, 7),
        61: (165, 181), 62: (52, 46), 63: (42, 24), 64: (100, 111),
        65: (99, 14), 66: (25, 93), 67: (56, 93), 68: (32, 19),
        69: (53, 81), 70: (10, 14), 71: (74, 77), 72: (79, 158),
        73: (87, 136), 74: (3, 145), 75: (122, 67), 76: (15, 91),
        77: (179, 82), 78: (102, 122), 79: (71, 71), 80: (158, 165),
        81: (155, 130), 82: (48, 134), 83: (29, 187), 84: (136, 145),
        85: (186, 116), 86: (77, 58), 87: (174, 11), 88: (18, 139),
        89: (171, 169), 90: (34, 108), 91: (61, 28), 92: (4, 95),
        93: (137, 30), 94: (8, 17), 95: (146, 60),
    }  #  fmt: skip

    def test_curve_repr(self) -> None:
        expected = "y^2 = x^3 - 4x + 0 (mod 191)"
        name = f"{self.curve}"
        assert name == expected

    def test_P_setup(self) -> None:
        P = self.curve.point(self.Px, self.Py)
        exp_P = 3, 46

        assert P.x == exp_P[0]
        assert P.y == exp_P[1]

    def test_compute_y(self) -> None:
        Qyy = self.curve.compute_y(self.Qx)
        assert Qyy is not None

        y0, y1 = Qyy

        Q0 = self.curve.point(self.Qx, y0)
        assert Q0.on_curve()

        Q1 = self.curve.point(self.Qx, y1)
        assert Q1.on_curve()

        assert (y0 + y1) % self.curve.p == 0

    def test_order(self) -> None:
        assert self.curve.order == 96

    def test_sums_on_curve(self) -> None:
        c = self.curve

        y = c.compute_y(self.Px)
        if not y:
            pytest.fail("failed to compute Py")
        else:
            Py = y[0]

        y = c.compute_y(self.Qx)
        if not y:
            pytest.fail("failed to compute Qy")
        else:
            Qy = y[0]

        P = c.point(self.Px, Py)
        Q = c.point(self.Qx, Qy)

        PpQ = P.add(Q)
        assert PpQ.on_curve()

        P2 = P.double()
        assert P2.on_curve()

    def test_generator(self) -> None:
        assert self.G.x == 146
        assert self.G.y == 131
        assert self.G.is_zero is False

        assert self.G.on_curve() is True

        for d in range(2, self.curve.order):
            dG = self.G.scaler_multiply(d)
            assert dG.on_curve()

    def test_double(self) -> None:
        G2 = self.G.double()
        assert (G2.x, G2.y) == (8, 174)

        G4 = G2.double()
        assert (G4.x, G4.y) == (4, 96)

    def test_doubling_only(self) -> None:
        doubled_only = [d for d in self.sage_vectors if d.bit_count() == 1]
        for d in doubled_only:
            x, y = self.sage_vectors[d]
            dG = self.G.scaler_multiply(d)
            assert dG.x == x
            assert dG.y == y

    def test_scaler_with_add(self) -> None:
        doubled_only = [d for d in self.sage_vectors if d.bit_count() != 1]
        for d in doubled_only:
            x, y = self.sage_vectors[d]
            dG = self.G.scaler_multiply(d)
            assert dG.x == x
            assert dG.y == y

    def test_pai(self) -> None:
        c = self.curve
        PaI = c.PAI
        P = c.point(self.Px, self.Py)
        negP = -P

        assert PaI.is_zero
        assert PaI + P == P
        assert P + PaI == P
        assert PaI == P - P
        assert negP + P == PaI

    def test_pai_immutable(self) -> None:
        c = self.curve
        Z = c.PAI
        P = c.point(self.Px, self.Py)
        with pytest.raises(NotImplementedError):
            Z += P

    def test_point_validation(self) -> None:
        with pytest.raises(ValueError):
            self.curve.point(self.Px, self.Py + 1)


def test_curve_validation() -> None:
    with pytest.raises(ValueError):
        Curve(2, 3, p=Modulus(31 * 73))

    with pytest.raises(ValueError):
        Curve(2, 3, p=Modulus(5))


if __name__ == "__main__":
    sys.exit(pytest.main(args=[__file__]))



"""Utility functions"""

from collections.abc import Iterator
import itertools
from hashlib import blake2b
from base64 import a85encode
from typing import Optional, Self
from toy_crypto.types import Byte


def digit_count(n: int, base: int = 10) -> int:
    """returns the number of digits (base b) of integer n.

    :raises ValueError: if base < 2
    :raises TypeError: if n is not an integer

    """

    if base < 2:
        raise ValueError("base must be at least 2")

    if not isinstance(n, int):
        raise TypeError("n must be an integer")

    n = abs(n)

    # Handling this case separately seems better than trying
    # to simulate a do-while in Python.
    if n == 0:
        return 1

    digits = 0
    while n > 0:
        digits += 1
        n = n // base
    return digits


class Xor:
    """Iterator that spits out xor of message with (repeated) pad.

    The iterator will run through successful bytes of message
    xor-ing those with successive bytes of pad, repeating
    pad if pad is shorter than message.

    Each iteration returns a non-negative int less than 256.
    """

    def __init__(
        self,
        message: Iterator[Byte] | bytes,
        pad: bytes,
    ) -> None:
        # Convert message to Iterator if needed
        self._message: Iterator[Byte] = iter(message)
        self._pad: Iterator[Byte] = itertools.cycle(pad)

    def __next__(self) -> Byte:
        b, p = next(zip(self._message, self._pad))
        return Byte(b ^ p)

    def __iter__(self: Self) -> Self:
        return self


def xor(message: bytes | Iterator[Byte], pad: bytes) -> bytes:
    """Returns the xor of message with a (repeated) pad.

    The pad is repeated if it is shorter than m.
    This can be thought of as bytewise Vigenère.
    """
    return bytes([b for b in Xor(message, pad)])


class Rsa129:
    """Text encoder/decoder used in RSA-129 challenge.

    Encoding scheme from Martin Gardner's 1977 article.
    """

    _abc: list[str] = list(" ABCDEFGHIJKLMNOPQRSTUVWXYZ")
    _abc_index: Optional[dict[str, int]] = None

    @classmethod
    def _make_index(cls) -> dict[str, int]:
        if cls._abc_index is None:
            cls._abc_index = {c: i for i, c in enumerate(cls._abc)}
        return cls._abc_index

    def __init__(self) -> None:
        self._make_index()

    @classmethod
    def encode(cls, text: str) -> int:
        """Encode text"""

        indx = cls._make_index()
        result = 0
        for c in text:
            result *= 100
            result += indx[c]
        return result

    @classmethod
    def decode(cls, number: int) -> str:
        """Decode text."""
        chars: list[str] = []
        while True:
            number, rem = divmod(number, 100)
            chars.append(cls._abc[rem])
            if number == 0:
                break
        return "".join(reversed(chars))


def hash_bytes(b: bytes) -> str:
    """Returns a python hashable from bytes.

    Primary intent is to have something that can be used
    as dictionary keys or members of sets.
    Collision resistance is the only security property
    that should be assumed here.

    The scheme may change from version to version
    """

    h = blake2b(b, digest_size=32).digest()
    t = str(a85encode(h))
    return t



# SPDX-FileCopyrightText: 2024-present Jeffrey Goldberg <jeffrey@goldmark.org>
#
# SPDX-License-Identifier: MIT

"""
Redundant definitions

Things here have been replace by what is in the nt module,
which wraps other, better implementations (mostly from primefac and math)
"""

import random  # random is good enough for Miller-Rabin.

from . import nt, types

# Generics require Python 3.12+, and it is hard to please type checkers here,
# So I am getting rid of this.

# E731 tells me to def prod instead of bind it to a lambda.
# https://docs.astral.sh/ruff/rules/lambda-assignment/
# def prod[T](iterable: Iterable[T]) -> T:  # type: ignore
#    """Returns the product of the elements of it"""
#    return reduce(lambda a, b: a * b, iterable)


# primes under 2^21
LOW_PRIMES: list[int] = [
    2,
    3,
    5,
    7,
    11,
    13,
    17,
    19,
    23,
    29,
    31,
    37,
    41,
    43,
    47,
    53,
    59,
    61,
    67,
    71,
    73,
    79,
    83,
    89,
    97,
    101,
    103,
    107,
    109,
    113,
    127,
    131,
    137,
    139,
    149,
    151,
    157,
    163,
    167,
    173,
    179,
    181,
    191,
    193,
    197,
    199,
    211,
    223,
    227,
    229,
    233,
    239,
    241,
    251,
    257,
    263,
    269,
    271,
    277,
    281,
    283,
    293,
    307,
    311,
    313,
    317,
    331,
    337,
    347,
    349,
    353,
    359,
    367,
    373,
    379,
    383,
    389,
    397,
    401,
    409,
    419,
    421,
    431,
    433,
    439,
    443,
    449,
    457,
    461,
    463,
    467,
    479,
    487,
    491,
    499,
    503,
    509,
    521,
    523,
    541,
    547,
    557,
    563,
    569,
    571,
    577,
    587,
    593,
    599,
    601,
    607,
    613,
    617,
    619,
    631,
    641,
    643,
    647,
    653,
    659,
    661,
    673,
    677,
    683,
    691,
    701,
    709,
    719,
    727,
    733,
    739,
    743,
    751,
    757,
    761,
    769,
    773,
    787,
    797,
    809,
    811,
    821,
    823,
    827,
    829,
    839,
    853,
    857,
    859,
    863,
    877,
    881,
    883,
    887,
    907,
    911,
    919,
    929,
    937,
    941,
    947,
    953,
    967,
    971,
    977,
    983,
    991,
    997,
    1009,
    1013,
    1019,
    1021,
    1031,
    1033,
    1039,
    1049,
    1051,
    1061,
    1063,
    1069,
    1087,
    1091,
    1093,
    1097,
    1103,
    1109,
    1117,
    1123,
    1129,
    1151,
    1153,
    1163,
    1171,
    1181,
    1187,
    1193,
    1201,
    1213,
    1217,
    1223,
    1229,
    1231,
    1237,
    1249,
    1259,
    1277,
    1279,
    1283,
    1289,
    1291,
    1297,
    1301,
    1303,
    1307,
    1319,
    1321,
    1327,
    1361,
    1367,
    1373,
    1381,
    1399,
    1409,
    1423,
    1427,
    1429,
    1433,
    1439,
    1447,
    1451,
    1453,
    1459,
    1471,
    1481,
    1483,
    1487,
    1489,
    1493,
    1499,
    1511,
    1523,
    1531,
    1543,
    1549,
    1553,
    1559,
    1567,
    1571,
    1579,
    1583,
    1597,
    1601,
    1607,
    1609,
    1613,
    1619,
    1621,
    1627,
    1637,
    1657,
    1663,
    1667,
    1669,
    1693,
    1697,
    1699,
    1709,
    1721,
    1723,
    1733,
    1741,
    1747,
    1753,
    1759,
    1777,
    1783,
    1787,
    1789,
    1801,
    1811,
    1823,
    1831,
    1847,
    1861,
    1867,
    1871,
    1873,
    1877,
    1879,
    1889,
    1901,
    1907,
    1913,
    1931,
    1933,
    1949,
    1951,
    1973,
    1979,
    1987,
    1993,
    1997,
    1999,
    2003,
    2011,
    2017,
    2027,
    2029,
    2039,
]

MAX_LOW_PRIME_SQUARE = LOW_PRIMES[-1] * LOW_PRIMES[-1]


def factor(n: int, ith: int = 0) -> nt.FactorList:
    """
    Returns list (prime, exponent) factors of n.
    Starts trial div at ith prime.

    This uses trial division by low primes for things upto the square
    of the largeest low primes and uses Fermet's method for everything else.
    Fermat's method is really bad when the
    factors are far away from each other, so my low primes list
    is longer than would make sense for other factoring strategies
    """

    if not types.is_positive_int(n):
        raise TypeError("input must be a positive int")

    if n == 1:
        return nt.FactorList([])

    factors = nt.FactorList()
    if n in LOW_PRIMES:
        return nt.FactorList([(n, 1)])

    root_n = isqrt(n)
    if root_n * root_n == n:
        return factor(root_n).pow(2)
    top = root_n

    if ith < len(LOW_PRIMES):
        for ith, p in enumerate(LOW_PRIMES[ith:]):
            if p > top:
                break
            reduced, remainder = divmod(n, p)
            if remainder == 0:
                exponent = 0
                prev_reduced = reduced
                while remainder == 0:
                    exponent += 1
                    prev_reduced = reduced
                    reduced, remainder = divmod(reduced, p)
                factors.append((p, exponent))
                return factors + factor(prev_reduced, ith)

    # n is not divisible by any of our low primes
    if n <= LOW_PRIMES[-1] ** 2:  # n is prime
        return nt.FactorList([(n, 1)])

    # Now we use Fermat's method (in the form of OLF
    # Note that OLF finds a (possibly composite factor),
    # So we will need to recurse and combine results.

    # OLE is a really, really, really slow way to test primality.
    # So we will do Miller-Rabin first
    if probably_prime(n):
        return nt.FactorList([(n, 1)])

    f = OLF(n)
    if f == 1:  # n is prime
        # This shouldn't happen, as we've already checked for primality
        # But I am leaving it in in case I somehow remove the
        # earlier primality check.
        return nt.FactorList([(n, 1)])

    return factor(f) + factor(n // f)


def gcd(a: int, b: int) -> int:
    """Returns greatest common denomenator of a and b."""
    while a != 0:
        a, b = b % a, a
    return b


def egcd(a: int, b: int) -> tuple[int, int, int]:
    """returns (g, x, y) such that a*x + b*y = gcd(a, b) = g."""
    x0, x1, y0, y1 = 0, 1, 1, 0
    while a != 0:
        (q, a), b = divmod(b, a), a
        y0, y1 = y1, y0 - q * y1
        x0, x1 = x1, x0 - q * x1
    return b, x0, y0


def modinv(a: int, m: int) -> int:
    """returns x such that a * x mod m = 1,"""
    g, x, _ = egcd(a, m)
    if g != 1:
        raise ValueError(f"{a} and {m} are not co-prime")
    return x % m


# from https://en.wikipedia.org/wiki/Integer_square_root
def isqrt(s: int) -> int:
    """returns the greatest n such that n * n =< s"""
    if s < 0:
        raise ValueError("s cannot be negative")
    if not isinstance(s, int):
        raise TypeError("s must be an int")

    if s < 2:
        return s

    x0: int = pow(2, (s.bit_length() // 2) + 1)
    x1: int = (x0 + s // x0) // 2

    while x1 < x0:
        x0 = x1
        x1 = (x0 + s // x0) // 2

    return x0


def is_square(n: int) -> bool:
    """True iff n is a perfect square."""

    # We can eliminate some quickly as
    #  - even squares must be 0 mod 4
    #  - odd squares must be 1 mod 8
    if n % 2 == 0:
        if n % 4 != 0:
            return False
    elif n % 8 != 1:
        return False

    r = isqrt(n)
    return r * r == n


# From https://programmingpraxis.com/2014/01/28/harts-one-line-factoring-algorithm/
def OLF(n: int) -> int:
    """Returns 1 if n is prime, else a factor (possibly composite) of n"""
    for ni in range(n, n * n, n):
        cs = isqrt(ni)
        # isqrt(x) gives us floor(sqrt(x)), but we need ceil(sqrt(n))
        if cs * cs != ni:
            cs += 1
        pcs = pow(cs, 2, n)
        if is_square(pcs):
            return gcd(n, cs - isqrt(pcs))

    # This will never be reached, but linters don't know that
    return 1


# lifted from https://gist.github.com/Ayrx/5884790
# k of 40 seems really high to me, but I see that the recommendation is from
# Thomas Pornin, so I am going to use that.
def probably_prime(n: int, k: int = 40) -> bool:
    """Returns True if n is prime or if you had really bad luck.

    Runs the Miller-Rabin primality test with k trials.
    Default value of k=40 is appropriate for use in key generation,
    but may be way high in other contexts.

    If you need a real primality check, use sympy.isprime() instead.
    """

    if n == 2:
        return True

    if n in LOW_PRIMES:
        return True
    if n <= MAX_LOW_PRIME_SQUARE:
        for p in LOW_PRIMES:
            if n % p == 0:
                return False
        return True

    # This s reduction is what distinguishes M-R from FLT
    r, s = 0, n - 1
    while s % 2 == 0:
        r += 1
        s //= 2

    # Now we use FLT for the reduced s, but still mod n
    for _ in range(k):
        a = random.randrange(2, n - 1)
        x = pow(a, s, n)
        if x == 1 or x == n - 1:
            # This a is good. Call the next witness!
            continue

        # square x for each time we reduced s for a passing x
        for _ in range(r - 1):
            x = pow(x, 2, n)
            if x == n - 1:
                # Our a really was good, once we tested squares of x
                break
        else:
            # We've exited the loop without finding a to be good
            return False

    # We've run all k trials, without any a telling us n is composite
    return True


def mod_sqrt(a: int, m: int) -> list[int]:
    """For odd prime m return (r, m-r) s.t. r^2 = a (mod m) if r exists.

    m must be an odd prime. If p is not prime, you might get a nice error,
    but sometimes you will get garbage results.
    """

    # I would have thought that there was a library for this.
    # Can't find it outside of sage, which is a bit big to import.

    # Algorithm taken from https://www.rieselprime.de/ziki/Modular_square_root

    if m < 3:
        raise ValueError("modulus must be an odd prime")

    if a == 1:
        return [1, m - 1]

    if m == 3:
        return []

    a = a % m
    if a == 0:
        return [0]

    # Use a small k, as we aren't claiming a definitive test
    if not probably_prime(m, k=5):
        raise ValueError("m is not prime")

    # check that a is a quadratic residue, return None if not
    if pow(a, (m - 1) // 2, m) != 1:
        return []

    if m % 4 == 3:  # if r exists then r = a^{(m+1)/2}
        r = pow(a, (m + 1) // 4, m)
        return [r, (m - r) % m]

    # Now we move to the tricky cases
    if m % 8 == 5:
        v = pow(2 * a, (m - 5) // 8, m)
        i = (2 * a * v * v) % m
        r = (a * v * (i - 1)) % m
        return [r, (m - r) % m]

    if m % 8 != 1:
        raise ValueError("modulus must be prime")

    # Now we are at the m % 8 == 1 case for which the algorithm is messy
    # Instructions quoted from
    #     https://www.rieselprime.de/ziki/Modular_square_root

    # Step 1: "Set e and odd q s.t. m = (2^e)q + 1"
    # Well, thanks. Let's figure out how to do that.

    # m is odd so we know that (2^1)q + 1 = m has
    # an integer q. So we start there
    e = 1
    q = (m - 1) // 2
    # while q is even if cut it in half and increment e
    while q % 2 == 0:
        e += 1
        q //= 2  # could do q >>= 1, but compiler will pick that if faster

    # just a check that we got an e and q
    if m != (2**e) * q + 1:
        raise Exception(f"Shouldn't happen: 2^{e} * {q} + 1 != {m}")
    if q % 2 == 0:
        raise Exception(f"Shouldn't happen: q ({q}) isn't odd.")

    # Step 2.
    # Find an x such that x is a quadratic non-residue of m.
    # We don't need good randomness, and half of possible x
    # are quadratic non-residues
    x: int = random.randint(2, m - 1)
    z: int = pow(x, q, m)
    while pow(z, 2 ** (e - 1), m) == 1:
        x = random.randint(2, m - 1)
        z = pow(x, q, m)

    # step 3:
    # x should now be a QNR in Zm
    y = z
    r = e
    x = pow(a, (q - 1) // 2, m)
    v = (a * x) % m
    w = (v * x) % m

    # step 4:
    # I do not understand why this all works.
    while w != 1:
        # step 5: "Find the smallest value of k such that w^(2^k) % m == 1"
        k = 1
        w2 = w * w
        while w2 % m != 1:
            w2 *= w2
            k += 1
        # step 6:
        d = pow(y, 2 ** (r - k - 1), m)
        y = (d * d) % m
        r = k
        v = (d * v) % m
        w = (w * y) % m

    return [v, (m - v) % m]


def lcm(a: int, b: int) -> int:
    """Least common multiple"""
    return abs(a * b) // gcd(a, b)



import timeit
import json
import polars as pl

# import cProfile
from toy_crypto.sieve import BaSieve, IntSieve, SetSieve, Sievish

repetitions = 1


def sieve_count(s_class: Sievish, size: int) -> int:
    s_class._reset()
    s = s_class.from_size(size)
    return s.count


sizes: list[int] = [10**n for n in range(2, 7)]
sizes += [5 * (10**n) for n in range(1, 6)]

s_classes: list[str] = [f"{c.__name__}" for c in (BaSieve, IntSieve, SetSieve)]

results: dict[str, list[float] | list[int]] = {"size": sizes}

for sieve_type in s_classes:
    times: list[float] = []
    for size in sizes:
        stmt = f"sieve_count({sieve_type}, {size})"
        t = timeit.timeit(stmt=stmt, number=repetitions, globals=globals())
        times.append(t)
    results[sieve_type] = times

j = json.dumps(results, indent="\t")

jfilename = "timings.json"
with open(jfilename, "w", encoding="utf-8") as f:
    f.write(j)
print(j)

df = pl.DataFrame(results)
print(df)

cfilename = "timings.csv"
df.write_csv(cfilename)
